@@ -1,9 +1,11 @@
-is.error <- function(Zt, Zp) {
+is.error <- function(Zt, # A vector of true cluster membership.
+                     Zp) # A vector of predictive cluster membership.
+{
   error <- array(data = 0, dim = length(Zp))
 
   if (length(Zt) != 0) {
     zt <- as.numeric(names(sort(table(Zt), decreasing = TRUE)))
-    zp <- as.numeric(levels(Zp))
+    zp <- sort(unique(Zp))  
 
     for (i in 1:length(zt)) {
       Zti <- Zt[which(Zt == zt[i])]
