@@ -32,6 +32,9 @@ function(x,
   if (ncol < 1) {
     stop(sQuote("ncol"), " must be greater than 0!", call. = FALSE)
   }
+  
+  Zp <- as.numeric(levels(x@Zp))[x@Zp]
+  Zt <- as.numeric(levels(x@Zt))[x@Zt]    
 
   d <- ncol(x@Dataset)
 
@@ -55,31 +58,35 @@ function(x,
 
   par(oma = c(1 + 0.2, 0.2, 0.2, 0.2))
 
-  ey <- as.matrix(x@Dataset); ep <- as.numeric(x@Zp) - 1
+  ey <- as.matrix(x@Dataset); ep <- Zp - 1
   
-  error <- is.error(x@Zt, x@Zp)
+  error <- is.error(Zt, Zp)
   
   ramp <- colorRamp(colors = c("magenta", "blue", "cyan", "green", "yellow", "red"),
     space = "rgb",
     interpolate = "linear")
     
-  s <- length(levels(x@Zp))
+  unique.Zp <- unique(Zp); s <- length(unique.Zp); sort.unique.Zp <- sort(unique.Zp)     
   
-  zlim <- c(0, max(1, s - 1)); zmax <- zlim[2]
-      
+  zlim <- c(0, max(1, s - 1)); zmax <- zlim[2]    
+    
   plot.col <- rgb(ramp(ep / zmax), maxColorValue = 255);
   
-  legend.col <- rgb(ramp(zlim[1]:zlim[2] / zmax), maxColorValue = 255)
+  legend.col <- rgb(ramp((sort.unique.Zp - 1) / zmax), maxColorValue = 255)
   
-  legend.pch <- rep(plot.pch, s)  
+  legend.text <- as.character(sort.unique.Zp)
+  
+  legend.pch <- rep(plot.pch, s)
+  
+  which.error <- which(error == 1); which.not.error <- which(error != 1)  
   
   if (N > 0) {
     figno <- 0
 
     for (i in 1:(d - 1)) {
       for (j in (i + 1):d) {
-        plot(x = ey[which(error != 1), i],
-          y = ey[which(error != 1), j],
+        plot(x = ey[which.not.error, i],
+          y = ey[which.not.error, j],
           type = "p",
           main = "",
           sub = "",
@@ -87,14 +94,14 @@ function(x,
           ylab = "",
           xlim = range(ey[, i]),
           ylim = range(ey[, j]),
-          col = plot.col[which(error != 1)],
+          col = plot.col[which.not.error],
           axes = FALSE,
           lwd = 1,
           cex = plot.cex,
           pch = plot.pch)
           
-        points(x = ey[which(error == 1), i],
-          y = ey[which(error == 1), j],
+        points(x = ey[which.error, i],
+          y = ey[which.error, j],
           type = "p",
           main = "",
           sub = "",
@@ -146,7 +153,7 @@ function(x,
           
           plot(0, 0, type = "n", bty = "n", xaxt = "n", yaxt = "n")
           
-         .legendA(s = s, col = legend.col, pch = legend.pch, error = sum(error) != 0)
+         .legendA(s = s, text = legend.text, col = legend.col, pch = legend.pch, error = sum(error) != 0)
   
           par(mfrow = c(nrow, ncol),
             cex = cex,
@@ -169,8 +176,8 @@ function(x,
     }
   }
   else {
-    plot(x = ey[which(error != 1), 1],
-      y = ep[which(error != 1)] + 1,
+    plot(x = ey[which.not.error, 1],
+      y = ep[which.not.error] + 1,
       type = "p",
       main = "",
       sub = "",
@@ -178,14 +185,14 @@ function(x,
       ylab = "",
       xlim = range(ey[, 1]),
       ylim = range(ep + 1),      
-      col = plot.col[which(error != 1)],
+      col = plot.col[which.not.error],
       axes = FALSE,
       lwd = 1,
       cex = plot.cex,
       pch = plot.pch)
           
-    points(x = ey[which(error == 1), 1],
-      y = ep[which(error == 1)] + 1,
+    points(x = ey[which.error, 1],
+      y = ep[which.error] + 1,
       type = "p",
       main = "",
       sub = "",
@@ -234,7 +241,7 @@ function(x,
           
     plot(0, 0, type = "n", bty = "n", xaxt = "n", yaxt = "n")
           
-    .legendA(s = s, col = legend.col, pch = legend.pch, error = sum(error) != 0)
+    .legendA(s = s, text = legend.text, col = legend.col, pch = legend.pch, error = sum(error) != 0)    
       
     par(mfrow = c(nrow, ncol),
       cex = cex,
@@ -291,6 +298,9 @@ function(x,
   if (ncol < 1) {
     stop(sQuote("ncol"), " must be greater than 0!", call. = FALSE)
   }
+  
+  Zp <- as.numeric(levels(x@Zp))[x@Zp]
+  Zt <- as.numeric(levels(x@Zt))[x@Zt]  
 
   d <- ncol(x@Dataset)
 
@@ -314,31 +324,35 @@ function(x,
 
   par(oma = c(1 + 0.2, 0.2, 0.2, 0.2))
 
-  ey <- as.matrix(x@Dataset); ep <- as.numeric(x@Zp) - 1
+  ey <- as.matrix(x@Dataset); ep <- Zp - 1
   
-  error <- is.error(x@Zt, x@Zp)
+  error <- is.error(Zt, Zp)
   
   ramp <- colorRamp(colors = c("magenta", "blue", "cyan", "green", "yellow", "red"),
     space = "rgb",
     interpolate = "linear")
     
-  s <- length(levels(x@Zp))
+  unique.Zp <- unique(Zp); s <- length(unique.Zp); sort.unique.Zp <- sort(unique.Zp)     
   
   zlim <- c(0, max(1, s - 1)); zmax <- zlim[2]
       
   plot.col <- rgb(ramp(ep / zmax), maxColorValue = 255);
   
-  legend.col <- rgb(ramp(zlim[1]:zlim[2] / zmax), maxColorValue = 255)
+  legend.col <- rgb(ramp((sort.unique.Zp - 1) / zmax), maxColorValue = 255)
+  
+  legend.text <- as.character(sort.unique.Zp)
   
   legend.pch <- rep(plot.pch, s)
   
+  which.error <- which(error == 1); which.not.error <- which(error != 1)  
+  
   if (N > 0) {
     figno <- 0
 
     for (i in 1:(d - 1)) {
       for (j in (i + 1):d) {
-        plot(x = ey[which(error != 1), i],
-          y = ey[which(error != 1), j],
+        plot(x = ey[which.not.error, i],
+          y = ey[which.not.error, j],
           type = "p",
           main = "",
           sub = "",
@@ -346,14 +360,14 @@ function(x,
           ylab = "",
           xlim = range(ey[, i]),
           ylim = range(ey[, j]),
-          col = plot.col[which(error != 1)],
+          col = plot.col[which.not.error],
           axes = FALSE,
           lwd = 1,
           cex = plot.cex,
           pch = plot.pch)
           
-        points(x = ey[which(error == 1), i],
-          y = ey[which(error == 1), j],
+        points(x = ey[which.error, i],
+          y = ey[which.error, j],
           type = "p",
           main = "",
           sub = "",
@@ -405,7 +419,8 @@ function(x,
           
           plot(0, 0, type = "n", bty = "n", xaxt = "n", yaxt = "n")
           
-          .legendA(s = s, col = legend.col, pch = legend.pch, error = sum(error) != 0)
+          
+          .legendA(s = s, text = legend.text, col = legend.col, pch = legend.pch, error = sum(error) != 0)          
   
           par(mfrow = c(nrow, ncol),
             cex = cex,
@@ -428,8 +443,8 @@ function(x,
     }
   }
   else {
-    plot(x = ey[which(error != 1), 1],
-      y = ep[which(error != 1)] + 1,
+    plot(x = ey[which.not.error, 1],
+      y = ep[which.not.error] + 1,
       type = "p",
       main = "",
       sub = "",
@@ -437,14 +452,14 @@ function(x,
       ylab = "",
       xlim = range(ey[, 1]),
       ylim = range(ep + 1),      
-      col = plot.col[which(error != 1)],
+      col = plot.col[which.not.error],
       axes = FALSE,
       lwd = 1,
       cex = plot.cex,
       pch = plot.pch)
           
-    points(x = ey[which(error == 1), 1],
-      y = ep[which(error == 1)] + 1,
+    points(x = ey[which.error, 1],
+      y = ep[which.error] + 1,
       type = "p",
       main = "",
       sub = "",
@@ -492,8 +507,8 @@ function(x,
       new = TRUE)
           
     plot(0, 0, type = "n", bty = "n", xaxt = "n", yaxt = "n")
-          
-    .legendA(s = s, col = legend.col, pch = legend.pch, error = sum(error) != 0)
+    
+    .legendA(s = s, text = legend.text, col = legend.col, pch = legend.pch, error = sum(error) != 0)          
       
     par(mfrow = c(nrow, ncol),
       cex = cex,
